#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/timer.h>
#include <sys/return_code.h>
#include <sys/process.h>
#include <cell/audio.h>
#include <cell/gcm.h>
#include <cell/sysmodule.h>
#include <cell/cell_fs.h>
#include <cell/pad.h>
#include <stddef.h>
#include <math.h>
#include <sysutil/sysutil_sysparam.h>
#include "snes9x.h"
#include "memmap.h"
#include "apu/apu.h"
#include "gfx.h"
#include "snapshot.h"
#include "controls.h"
#include "cheats.h"
#include "movie.h"
#include "logger.h"
#include "display.h"
#include "conffile.h"
#include "font.h"

SYS_PROCESS_PARAM(1001, 0x10000);

using namespace cell::Gcm;

static uint32_t local_heap_next;
static void *host_io_buffer;
static bool is_running;
static int mode_switch = 0;
static int rom_loaded_count = 0;

typedef struct {
	void *address;
	uint32_t offset;
	uint32_t pitch;
	uint32_t width;
	uint32_t height;
} color_buf_t;

typedef struct {
	uint8_t a;
	uint8_t r;
	uint8_t g;
	uint8_t b;
} color4_t;

#define MIN(x,y) ((x) < (y) ? (x) : (y))
#define MAX(x,y) ((x) > (y) ? (x) : (y))

#define CMD_BUFFER_SIZE (1024 * 64)
#define IO_BUFFER_SIZE (1024 * 1024)
#define COLOR_BUFFER_NUM 2
#define SHADER(x) extern uint32_t _binary_##x##_start; static unsigned char *shader_##x##_ptr = (unsigned char *)&_binary_##x##_start; CGprogram shader_##x; static void *shader_##x##_ucode; static uint32_t shader_##x##_offset


#define USRDIR "/dev_hdd0/game/SNES90000/USRDIR"

SHADER(fpshader_fpo);

static color4_t *display;
static color4_t *display_back;
static uint32_t frame_index;
static color_buf_t color_buf[COLOR_BUFFER_NUM];
uint32_t depth_pitch, depth_offset, display_offset, display_back_offset;
static const char *debug_text;
static char rom_path[1024];
static const uint32_t dma_label_id = 128;
static uint32_t *dma_label;
static uint32_t dma_label_value;


void *localHeapAlloc(const uint32_t size) {
	uint32_t allocated_size = (size + 1023) & (~1023);
	void *addr = (void *)local_heap_next;

	local_heap_next += allocated_size;
	return addr;
}

void *localHeapAlign(const uint32_t align, const uint32_t size) {
	local_heap_next = (local_heap_next + align - 1) & (~(align - 1));
	return localHeapAlloc(size);
}

static void setRenderTarget(const uint32_t Index)
{
	CellGcmSurface sf;
	//set to same value as S9xPixelFormat
	//sf.colorFormat 	= CELL_GCM_SURFACE_A8R8G8B8;
	sf.colorFormat = CELL_GCM_SURFACE_R5G6B5;
	sf.colorTarget	= CELL_GCM_SURFACE_TARGET_0;
	sf.colorLocation[0]	= CELL_GCM_LOCATION_LOCAL;
	sf.colorOffset[0] 	= color_buf[Index].offset;
	sf.colorPitch[0] 	= color_buf[Index].pitch;

	sf.colorLocation[1]	= CELL_GCM_LOCATION_LOCAL;
	sf.colorLocation[2]	= CELL_GCM_LOCATION_LOCAL;
	sf.colorLocation[3]	= CELL_GCM_LOCATION_LOCAL;
	sf.colorOffset[1] 	= 0;
	sf.colorOffset[2] 	= 0;
	sf.colorOffset[3] 	= 0;
	sf.colorPitch[1]	= 64;
	sf.colorPitch[2]	= 64;
	sf.colorPitch[3]	= 64;

	sf.depthFormat 	= CELL_GCM_SURFACE_Z24S8;
	sf.depthLocation	= CELL_GCM_LOCATION_LOCAL;
	sf.depthOffset	= depth_offset;
	sf.depthPitch 	= depth_pitch;

	sf.type		= CELL_GCM_SURFACE_PITCH;
	sf.antialias	= CELL_GCM_SURFACE_CENTER_1;

	sf.width 		= color_buf[0].width;
	sf.height 		= color_buf[0].height;
	sf.x 		= 0;
	sf.y 		= 0;
	cellGcmSetSurface(&sf);
}


static void setDrawEnv(void)
{
	cellGcmSetColorMask(gCellGcmCurrentContext, CELL_GCM_COLOR_MASK_B|
			CELL_GCM_COLOR_MASK_G|
			CELL_GCM_COLOR_MASK_R|
			CELL_GCM_COLOR_MASK_A);

	cellGcmSetColorMaskMrt(gCellGcmCurrentContext, 0);
	uint16_t x,y,w,h;
	float min, max;
	float scale[4],offset[4];

	x = 0;
	y = 0;
	w = color_buf[0].width;
	h = color_buf[0].height;
	min = 0.0f;
	max = 1.0f;
	scale[0] = w * 0.5f;
	scale[1] = h * -0.5f;
	scale[2] = (max - min) * 0.5f;
	scale[3] = 0.0f;
	offset[0] = x + scale[0];
	offset[1] = h - y + scale[1];
	offset[2] = (max + min) * 0.5f;
	offset[3] = 0.0f;

	cellGcmSetViewport(gCellGcmCurrentContext, x, y, w, h, min, max, scale, offset);
	cellGcmSetClearColor(gCellGcmCurrentContext, (64<<0)|(64<<8)|(64<<16)|(64<<24));
}

static void init_display () {
	int32_t ret;
	CellVideoOutState video_state;
	CellVideoOutResolution resolution;
	CellVideoOutConfiguration video_config;
	CellGcmConfig gcm_config;
	uint32_t color_buffer_size;
	void *color_buffer_base;

	video_state.displayMode.resolutionId = CELL_VIDEO_OUT_RESOLUTION_480;
	video_state.displayMode.scanMode = CELL_VIDEO_OUT_SCAN_MODE_PROGRESSIVE;
	video_state.displayMode.refreshRates = CELL_VIDEO_OUT_REFRESH_RATE_59_94HZ;
	video_state.displayMode.conversion = CELL_VIDEO_OUT_DISPLAY_CONVERSION_NONE;
	video_state.displayMode.aspect = CELL_VIDEO_OUT_ASPECT_4_3;

	host_io_buffer = memalign(1024 * 1024, IO_BUFFER_SIZE);
	cellGcmInit(CMD_BUFFER_SIZE, IO_BUFFER_SIZE, host_io_buffer);

	ret = cellVideoOutGetState(CELL_VIDEO_OUT_PRIMARY, 0, &video_state);

	if (ret != CELL_OK)
		return;

	cellVideoOutGetResolution(video_state.displayMode.resolutionId, &resolution);

	depth_pitch = resolution.width * 4;

	memset(&video_config, 0, sizeof(CellVideoOutConfiguration));

	video_config.resolutionId = video_state.displayMode.resolutionId;
	video_config.format = CELL_VIDEO_OUT_BUFFER_COLOR_FORMAT_X8R8G8B8;
	video_config.aspect = video_state.displayMode.aspect;
	video_config.pitch = resolution.width * 4;

	cellVideoOutConfigure(CELL_VIDEO_OUT_PRIMARY, &video_config, NULL, 0);
	cellGcmSetFlipMode(CELL_GCM_DISPLAY_VSYNC);

	cellGcmGetConfiguration(&gcm_config);
	local_heap_next = (uint32_t)gcm_config.localAddress;

	color_buffer_size = resolution.width * resolution.height * 4;
	color_buffer_base = localHeapAlign(16, color_buffer_size * COLOR_BUFFER_NUM);
	memset(color_buffer_base, 0, color_buffer_size);

	for (int i = 0; i < COLOR_BUFFER_NUM; ++i) {
		void *buf_addr = (void *)((uint32_t)color_buffer_base + (i * color_buffer_size));

		color_buf[i].address = buf_addr;
		color_buf[i].width = resolution.width;
		color_buf[i].height = resolution.height;
		color_buf[i].pitch = resolution.width * 4;

		cellGcmAddressToOffset(buf_addr, &color_buf[i].offset);
		cellGcmSetDisplayBuffer(i, color_buf[i].offset, color_buf[i].pitch, color_buf[i].width, color_buf[i].height);
	}

	void *depth_addr = localHeapAlign(16, depth_pitch * resolution.height);
	cellGcmAddressToOffset(depth_addr, &depth_offset);
}

static void init_shader () {
	shader_fpshader_fpo = (CGprogram)shader_fpshader_fpo_ptr;
	cellGcmCgInitProgram(shader_fpshader_fpo);

	uint32_t ucode_size;
	void *ucode;
	cellGcmCgGetUCode(shader_fpshader_fpo, &ucode, &ucode_size);
	shader_fpshader_fpo_ucode = localHeapAlign(128, ucode_size);
	memcpy(shader_fpshader_fpo_ucode, ucode, ucode_size);
	cellGcmAddressToOffset(shader_fpshader_fpo_ucode, &shader_fpshader_fpo_offset);
}

void set_render_state () {
	cellGcmSetFragmentProgram(shader_fpshader_fpo, shader_fpshader_fpo_offset);
}

void flip () {
	static bool first = true;

	cellGcmFlush();

	if (!first) {
		while (cellGcmGetFlipStatus() != 0 || *dma_label != dma_label_value)
			sys_timer_usleep(300);
	}

	cellGcmResetFlipStatus();

	if (cellGcmSetFlip(frame_index) != CELL_OK) return;
	cellGcmFlush();
	
	setDrawEnv();
	set_render_state();

	cellGcmSetWaitFlip();

	frame_index = (frame_index + 1) % COLOR_BUFFER_NUM;
	setRenderTarget(frame_index);

	first = false;
}

void sysutil_exit_callback (uint64_t status, uint64_t param, void *userdata) {
	(void) param;
	(void) userdata;

	switch (status) {
		case CELL_SYSUTIL_REQUEST_EXITGAME:
			is_running = false;
			mode_switch = 2;
			break;
		case CELL_SYSUTIL_DRAWING_BEGIN:
		case CELL_SYSUTIL_DRAWING_END:
			break;
	}
}

void _splitpath (const char *path, char *drive, char *dir, char *fname, char *ext)
{
	*drive = 0;

	const char	*slash = strrchr(path, SLASH_CHAR),
				*dot   = strrchr(path, '.');

	if (dot && slash && dot < slash)
		dot = NULL;

	if (!slash)
	{
		*dir = 0;

		strcpy(fname, path);

		if (dot)
		{
			fname[dot - path] = 0;
			strcpy(ext, dot + 1);
		}
		else
			*ext = 0;
	}
	else
	{
		strcpy(dir, path);
		dir[slash - path] = 0;

		strcpy(fname, slash + 1);

		if (dot)
		{
			fname[dot - slash - 1] = 0;
			strcpy(ext, dot + 1);
		}
		else
			*ext = 0;
	}
}

void _makepath (char *path, const char *, const char *dir, const char *fname, const char *ext)
{
	if (dir && *dir)
	{
		strcpy(path, dir);
		strcat(path, SLASH_STR);
	}
	else
		*path = 0;

	strcat(path, fname);

	if (ext && *ext)
	{
		strcat(path, ".");
		strcat(path, ext);
	}
}

void S9xParsePortConfig (ConfigFile &conf, int pass) {
}

void S9xExtraUsage (void) {
}

void S9xParseArg (char **argv, int &i, int argc) {
}

bool S9xPollAxis (uint32_t id, int16_t *value) {
	return false;
}

bool S9xPollButton (uint32_t id, bool *pressed) {
	return false;
}

bool S9xPollPointer (uint32_t id, int16_t *x, int16_t *y) {
	return false;
}

void S9xHandlePortCommand (s9xcommand_t cmd, int16_t data1, int16_t data2) {
}

void S9xCloseSnapshotFile (STREAM file) {
}

const char *S9xStringInput (const char *message) {
	return NULL;
}

bool8 S9xOpenSnapshotFile (const char *filepath, bool8 read_only, STREAM *file) {
	return false;
}

void S9xExit (void) {
}

bool8 S9xInitUpdate (void) {
	return true;
}

bool8 S9xDeinitUpdate (int width, int height) {
	int baseX = 200, baseY = 200, xscale, yscale, destW, destH, xinc = 0, yinc = 0;
	static char buf[1024] = { '\0' };
	uint16_t *gfxptr;
       	uint32_t *displayptr;

	xscale = color_buf[0].width / width;
	yscale = color_buf[0].height / height;

	if (xscale > yscale)
		xscale = yscale;
	else if (yscale > xscale)
		yscale = xscale;

	destW = width * xscale;
	destH = height * yscale;

	baseX = color_buf[0].width/2 - destW/2;
	baseY = color_buf[0].height/2 - destH/2;

	displayptr = (uint32_t *)&display[baseY * color_buf[0].width + baseX];
	gfxptr = GFX.Screen;

	for (int y = 0; y < destH; ++y) {
		for (int x = 0; x < destW; ++x) {
			*(displayptr++) = ((*gfxptr & 0x1f) << 3) | ((*gfxptr & 0x7e0) << 5) | ((*gfxptr & 0xf800) << 8);

			if (++xinc == xscale) {
				gfxptr++;
				xinc = 0;
			}
		}

		if (++yinc == yscale) {
			gfxptr += (GFX.Pitch>>1) - width;
			yinc = 0;
		} else {
			gfxptr -= width;
		}

		displayptr += color_buf[0].width - destW;
	}

	return true;
}

bool8 S9xContinueUpdate (int width, int height) {
	return true;
}

void S9xMessage (int type, int number, char const *message) {
}

bool8 S9xOpenSoundDevice (void) {
	return true;
}

const char * S9xGetFilename (const char *ex, enum s9x_getdirtype dirtype)
{
	static char	s[PATH_MAX + 1];
	char		drive[_MAX_DRIVE + 1], dir[_MAX_DIR + 1], fname[_MAX_FNAME + 1], ext[_MAX_EXT + 1];

	_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
	snprintf(s, PATH_MAX + 1, "%s%s%s%s", S9xGetDirectory(dirtype), SLASH_STR, fname, ex);

	return (s);
}

const char * S9xGetFilenameInc (const char *ex, enum s9x_getdirtype dirtype)
{
	static char	s[PATH_MAX + 1];
	char		drive[_MAX_DRIVE + 1], dir[_MAX_DIR + 1], fname[_MAX_FNAME + 1], ext[_MAX_EXT + 1];

	unsigned int	i = 0;
	const char		*d;

	_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
	d = S9xGetDirectory(dirtype);

	snprintf(s, PATH_MAX + 1, "%s%s%s.%03d%s", d, SLASH_STR, fname, i++, ex);

	return (s);
}

const char * S9xBasename (const char *f)
{
	const char	*p;

	if ((p = strrchr(f, '/')) != NULL || (p = strrchr(f, '\\')) != NULL)
		return (p + 1);

	return (f);
}

const char *S9xGetDirectory (enum s9x_getdirtype dirtype) {
	static char s[PATH_MAX + 1];

	switch (dirtype) {
		case SRAM_DIR:
		case ROMFILENAME_DIR:
			strncpy(s, Memory.ROMFilename, PATH_MAX + 1);
			s[PATH_MAX] = 0;

			for (int i = strlen(s); i >= 0; --i) {
				if (s[i] == SLASH_CHAR) {
					s[i] = 0;
					break;
				}
			}

			return s;
		default:
			return USRDIR;
	}
}

const char *S9xChooseFilename (bool8 read_only) {
	return NULL;
}

const char *S9xChooseMovieFilename (bool8 read_only) {
	return NULL;
}

void S9xAutoSaveSRAM (void) {
	Memory.SaveSRAM(S9xGetFilename(".srm", SRAM_DIR));
}

void S9xToggleSoundChannel (int c) {
}

void S9xSetPalette (void) {
}

void S9xSyncSpeed (void) {
}

// SNES buttons

#define FONT_HEIGHT 9
#define FONT_WIDTH 8
#define FONT_SCALE 1

#define BTN_A 1
#define BTN_B 2
#define BTN_X 3
#define BTN_Y 4
#define BTN_SELECT 5
#define BTN_START 6
#define BTN_L 7
#define BTN_R 8
#define BTN_LEFT 9
#define BTN_RIGHT 10
#define BTN_UP 11
#define BTN_DOWN 12

#define PAD_1 1
#define PAD_2 2

#define MAKE_BUTTON(pad, btn) (((pad)<<4)|(btn))

// PS3 buttons (pad data)
#define	CTRL_SELECT		(1<<0)
#define	CTRL_L3			(1<<1)
#define	CTRL_R3			(1<<2)
#define	CTRL_START		(1<<3)
#define	CTRL_UP			(1<<4)
#define	CTRL_RIGHT		(1<<5)
#define	CTRL_DOWN		(1<<6)
#define	CTRL_LEFT		(1<<7)
#define	CTRL_L2			(1<<8)
#define	CTRL_R2			(1<<9)
#define	CTRL_L1			(1<<10)
#define	CTRL_R1			(1<<11)
#define	CTRL_TRIANGLE	(1<<12)
#define	CTRL_CIRCLE		(1<<13)
#define	CTRL_CROSS		(1<<14)
#define	CTRL_SQUARE		(1<<15)

void debugOutChr (int baseX, int baseY, uint8_t chr, color4_t color) {
	color4_t black = { 0, 0, 0, 0 };

	int line = ((chr - 32) >> 4) * FONT_HEIGHT;
	int offset = ((chr - 32) & 15) * FONT_WIDTH;
	color4_t *ptr = &display[baseY * color_buf[0].width + baseX];

	for (int h = 0; h < FONT_HEIGHT * FONT_SCALE; ++h, ptr += color_buf[0].width - FONT_WIDTH * FONT_SCALE) {
		for (int w = 0; w < FONT_WIDTH * FONT_SCALE; ++w, ++ptr) {
			char p = font[line + h / FONT_SCALE][offset + w / FONT_SCALE];

			if (p == '#')
				*ptr = color;
			else if (p == '.')
				*ptr = black;
		}
	}
}

void debugOutStr (int baseX, int baseY, const char *str, color4_t color) {
	for (const char *ptr = str; *ptr; ++ptr) {
		debugOutChr(baseX, baseY, *ptr, color);
		baseX += FONT_WIDTH * FONT_SCALE;
	}
}

void debugOutStr (int baseX, int baseY, const char *str) {
	color4_t white = { 255, 255, 255, 255 };

	debugOutStr(baseX, baseY, str, white);
}

bool init_snes9x () {
	ZeroMemory(&Settings, sizeof(Settings));

	Settings.MouseMaster = TRUE;
	Settings.SuperScopeMaster = TRUE;
	Settings.JustifierMaster = TRUE;
	Settings.MultiPlayer5Master = TRUE;
	Settings.FrameTimePAL = 20000;
	Settings.FrameTimeNTSC = 16667;
	Settings.SixteenBitSound = TRUE;
	Settings.Stereo = TRUE;
	Settings.SoundPlaybackRate = 48000;
	Settings.SoundInputRate = 31942;
	Settings.SoundSync = TRUE;
	Settings.SupportHiRes = TRUE;
	Settings.Transparency = TRUE;
	Settings.AutoDisplayMessages = TRUE;
	Settings.InitialInfoStringTimeout = 120;
	Settings.HDMATimingHack = 100;
	Settings.BlockInvalidVRAMAccessMaster = TRUE;
	Settings.StopEmulation = TRUE;
	Settings.WrongMovieStateProtection = TRUE;
	Settings.DumpStreamsMaxFrames = -1;
	Settings.StretchScreenshots = 1;
	Settings.SnapshotScreenshots = TRUE;
	Settings.SkipFrames = AUTO_FRAMERATE;
	Settings.TurboSkipFrames = 15;
	Settings.CartAName[0] = 0;
	Settings.CartBName[0] = 0;
	Settings.AutoSaveDelay = 0;

	CPU.Flags = 0;

	Memory.Init();

	S9xInitAPU();

	S9xInitSound(100, 25);

	S9xSetSoundMute(TRUE);
	S9xSetRenderPixelFormat(RGB565);

	if (!Memory.LoadROM(rom_path)) {
		debug_text = "load rom failed";
		return false;
	}

	Memory.LoadSRAM(S9xGetFilename(".srm", SRAM_DIR));

	GFX.Pitch = 1024;
	GFX.Screen = (uint16_t *)memalign(1024, 1024 * 478);

	S9xGraphicsInit();

	// Controller initialization
	S9xSetController(0, CTL_JOYPAD, 0, 0, 0, 0);
	S9xSetController(1, CTL_JOYPAD, 1, 0, 0, 0);

	S9xUnmapAllControls();

#define MAP_BUTTON(id, name) S9xMapButton((id), S9xGetCommandT((name)), false)

	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_A), "Joypad1 A");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_B), "Joypad1 B");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_X), "Joypad1 X");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_Y), "Joypad1 Y");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_SELECT), "Joypad1 Select");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_START), "Joypad1 Start");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_L), "Joypad1 L");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_R), "Joypad1 R");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_LEFT), "Joypad1 Left");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_RIGHT), "Joypad1 Right");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_UP), "Joypad1 Up");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_DOWN), "Joypad1 Down");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_A), "Joypad2 A");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_B), "Joypad2 B");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_X), "Joypad2 X");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_Y), "Joypad2 Y");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_SELECT), "Joypad2 Select");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_START), "Joypad2 Start");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_L), "Joypad2 L");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_R), "Joypad2 R");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_LEFT), "Joypad2 Left");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_RIGHT), "Joypad2 Right");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_UP), "Joypad2 Up");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_DOWN), "Joypad2 Down");
	
	return true;

}

bool read_pad_buttons (uint32_t *result, int id = 0) {
	CellPadInfo info;
	CellPadData data;

	cellPadGetInfo(&info);

	if (info.status[id] == CELL_PAD_STATUS_CONNECTED) {
		cellPadGetData(id, &data);

		if (data.len == 0)
			return false;

		uint32_t buttons = (data.button[CELL_PAD_BTN_OFFSET_DIGITAL1] & 0xff) | ((data.button[CELL_PAD_BTN_OFFSET_DIGITAL2] & 0xFF) << 8);
		*result = buttons;
		return true;
	}

	return false;
}

void read_pad () {
	uint32_t buttons;

	for (int i = 0; i < 2; ++i) {
		if (read_pad_buttons(&buttons, i)) {
			int pad_id = i + 1;

			S9xReportButton(MAKE_BUTTON(pad_id, BTN_A), (buttons & CTRL_CIRCLE) != 0);
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_B), (buttons & CTRL_CROSS) != 0);
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_X), (buttons & CTRL_TRIANGLE) != 0);
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_Y), (buttons & CTRL_SQUARE) != 0);

			S9xReportButton(MAKE_BUTTON(pad_id, BTN_SELECT), (buttons & CTRL_SELECT) != 0);
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_START), (buttons & CTRL_START) != 0);

			S9xReportButton(MAKE_BUTTON(pad_id, BTN_L), (buttons & CTRL_L1) != 0);
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_R), (buttons & CTRL_R1) != 0);

			S9xReportButton(MAKE_BUTTON(pad_id, BTN_UP), (buttons & CTRL_UP) != 0);
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_DOWN), (buttons & CTRL_DOWN) != 0);
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_LEFT), (buttons & CTRL_LEFT) != 0);
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_RIGHT), (buttons & CTRL_RIGHT) != 0);
			if ((buttons & CTRL_R2) != 0 && (buttons & CTRL_L2) != 0)
			{
				is_running = false;
				rom_loaded_count++;
				if(rom_loaded_count == 4)
				{
					mode_switch = 2;
				}
				else
				{
					mode_switch = 0;
				}
			}
		}
	}
}

#define AUDIO_BLOCK_COUNT 16 
#define AUDIO_CHANNEL_COUNT 2

static uint32_t audio_port;
static float *audio_buffer;
static float *audio_buffer_end;
sys_addr_t audio_read_index_addr;
static int audio_block_index;
static int audio_last_read_block;

void init_sound () {
	CellAudioPortParam params;
	CellAudioPortConfig config;

	cellAudioInit();

	params.nChannel = AUDIO_CHANNEL_COUNT;
	params.nBlock = AUDIO_BLOCK_COUNT;
	params.attr = 0;

	if (cellAudioPortOpen(&params, &audio_port) != CELL_OK) {
		cellAudioQuit();
		return;
	}

	cellAudioGetPortConfig(audio_port, &config);
	audio_buffer = (float *)config.portAddr;
	audio_buffer_end = (float *)(config.portAddr + config.portSize);
	audio_read_index_addr = config.readIndexAddr;

	cellAudioPortStart(audio_port);
}

void make_noise () {
	int16_t samples[256 * 2 * 2];
	char dbuf[128];

	if (S9xGetSampleCount() < 512)
		return;

	int current_block = *(uint64_t *)audio_read_index_addr;

	if (current_block == audio_last_read_block)
		return;

	while (S9xGetSampleCount() >= 512 && audio_block_index != current_block) {
		S9xMixSamples((uint8 *)samples, 256 * 2);

		float *buf = audio_buffer + AUDIO_CHANNEL_COUNT * 256 * audio_block_index;
		for (int i = 0; i < 256 * AUDIO_CHANNEL_COUNT; ++i)
			buf[i] = ((float)samples[i])/32768.0f;

		audio_block_index = (audio_block_index + 1) % AUDIO_BLOCK_COUNT;
	}

	audio_last_read_block = current_block;
}

void render(bool sync = true) {
	int render_to = frame_index;

	cellSysutilCheckCallback();
	cellGcmSetClearSurface(CELL_GCM_CLEAR_Z | CELL_GCM_CLEAR_R | CELL_GCM_CLEAR_G | CELL_GCM_CLEAR_B | CELL_GCM_CLEAR_A);
	cellGcmSetTransferData(gCellGcmCurrentContext,
		       	CELL_GCM_TRANSFER_MAIN_TO_LOCAL, color_buf[render_to].offset, color_buf[render_to].pitch, 
			display_offset, color_buf[0].pitch, color_buf[0].pitch, color_buf[0].height);
	
	flip();

	if (sync) {
		cellGcmSetWriteBackEndLabel(dma_label_id, ++dma_label_value);
		cellGcmSetWaitLabel(dma_label_id, dma_label_value);
	}

	uint32_t display_offset_tmp = display_offset;
	color4_t *display_tmp = display;

	display = display_back;
	display_offset = display_back_offset;

	display_back = display_tmp;
	display_back_offset = display_offset_tmp;
}

void rom_menu () {
	bool needs_update = true;
	char *file_names[8192] = { '\0' };
	bool is_dir[8192] = { false };
	char path[1024];
	const char *msg = NULL;
	int fd;
	int num_per_page = 47, current_index = 0, file_count = 0;
	uint32_t buttons;

	strcpy(path, "/");

	while (is_running) {
		if (needs_update) {
			needs_update = false;
			current_index = 0;

			for (int i = 0; i < 8192; ++i) {
				if (file_names[i]) {
					free(file_names[i]);
					file_names[i] = NULL;
				} else {
					break;
				}
			}

			int i = 0;
			if (cellFsOpendir(path, &fd) == CELL_FS_SUCCEEDED) {
				uint64_t nread;
				CellFsDirent dirent;

				while (cellFsReaddir(fd, &dirent, &nread) == CELL_FS_SUCCEEDED) {
					if (nread == 0)
						break;

					if (dirent.d_type != CELL_FS_TYPE_DIRECTORY &&
					    dirent.d_type != CELL_FS_TYPE_REGULAR)
						continue; // so lazy.

					if (dirent.d_type == CELL_FS_TYPE_REGULAR) {
						int namelen = strlen(dirent.d_name);
						if (namelen < 4)
							continue;

						char *ext = dirent.d_name + namelen - 4;

						if (strcasecmp(ext, ".smc") && strcasecmp(ext, ".fig") && strcasecmp(ext, ".sfc"))
							continue;
					}

					is_dir[i] = dirent.d_type == CELL_FS_TYPE_DIRECTORY;
					file_names[i] = strdup(dirent.d_name);

					++i;

					if (i == 8192)
						break;
				}

				cellFsClosedir(fd);
			} else {
				msg = "Directory not found. Try again!";
			}

			file_count = i;

			if (file_count == 0)
				msg = "No ROMs. =(";

			needs_update = false;
		}


		if (read_pad_buttons(&buttons)) {
			if (buttons & CTRL_DOWN)
				current_index = MIN(current_index + 1, file_count - 1);

			if (buttons & CTRL_UP)
				current_index = MAX(current_index - 1, 0);

			if (buttons & CTRL_RIGHT)
				current_index = MIN(current_index + 5, file_count - 1);

			if (buttons & CTRL_LEFT)
				current_index = MAX(current_index - 5, 0);

			if (buttons & CTRL_R1)
				current_index = MIN(current_index + num_per_page, file_count - 1);

			if (buttons & CTRL_L1)
				current_index = MAX(current_index - num_per_page, 0);

			if (buttons & CTRL_CROSS) {
				if (is_dir[current_index]) {
					sprintf(path, "%s/%s", path, file_names[current_index]);
					needs_update = true;
					continue;
				} else {
					sprintf(rom_path, "%s/%s", path, file_names[current_index]);
					mode_switch = 1;
					return;
				}
			}
		}


		memset(display, 0, color_buf[0].pitch * color_buf[0].height);

		int page_number = current_index / num_per_page;
		int page_base = page_number * num_per_page;

		for (int i = page_base; i < file_count && i < page_base + num_per_page; ++i) {
			color4_t white = { 0xFF, 0xFF, 0xFF, 0xFF };
			color4_t red = { 0xFF, 0xFF, 0, 0 };
			color4_t green = { 0xFF, 0, 0xFF, 0 };

			debugOutStr(32, 32 + FONT_HEIGHT * FONT_SCALE * (i - page_base), file_names[i], 
					i == current_index ? red : is_dir[i] ? green : white);
		}

		if (msg)
			debugOutStr(32, 32 + FONT_HEIGHT * FONT_SCALE * (num_per_page + 1), msg);

		render();
	}
}

void start_snes9x()
{
	bool snes9x_loaded = false;
	if (!(snes9x_loaded = init_snes9x())) {
		Settings.Paused = TRUE;
	}

	int count = 0;
	
	init_sound();
	S9xSetSoundMute(FALSE);

	memset(display, 0, color_buf[0].height * color_buf[0].pitch);
	memset(display_back, 0, color_buf[0].height * color_buf[0].pitch);

	while (is_running) {
		read_pad();

		if (!Settings.Paused) {
			make_noise();
			if (S9xSyncSound())
				S9xMainLoop();
		}

		if (debug_text)
			debugOutStr(100, 100, debug_text);

		render(false);
	}


	if (snes9x_loaded)
		Memory.SaveSRAM(S9xGetFilename(".srm", SRAM_DIR));
	snes9x_loaded = false;
	is_running = true;

}

int main (void) {
	cellSysmoduleLoadModule(CELL_SYSMODULE_FS);
	cellSysmoduleLoadModule(CELL_SYSMODULE_IO);
	cellSysmoduleLoadModule(CELL_SYSMODULE_AUDIO);

	cellSysutilRegisterCallback(0, sysutil_exit_callback, NULL);

	cellPadInit(7);
	init_display();
	init_shader();

	dma_label = cellGcmGetLabelAddress(dma_label_id);
	*dma_label = dma_label_value;

	is_running = true;

	uint32_t buf_size = sizeof(color4_t) * color_buf[0].width * color_buf[0].height;
	buf_size = (buf_size + 1024 * 1024) & (~(1024 * 1024 - 1));

	color4_t *buf = (color4_t *)memalign(1024 * 1024, buf_size);
	cellGcmMapMainMemory(buf, buf_size, &display_back_offset);

	memset(buf, 0, buf_size);
	display_back = buf;

	buf = (color4_t *)memalign(1024 * 1024, buf_size);
	cellGcmMapMainMemory(buf, buf_size, &display_offset);

	memset(buf, 0, buf_size);
	display = buf;

	setDrawEnv();

	set_render_state();
	setRenderTarget(frame_index);

	while(1)
	{
		switch(mode_switch)
		{
			case 0:
				rom_menu();
				break;
			case 1:
				start_snes9x();
				break;
			case 2:
				return 0;
		}
	}

	return 0;
}
